"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.section import BaseSectionConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class BrainOverridesConfiguration(BaseSectionConfigurationData):

    def __init__(self):
        BaseSectionConfigurationData.__init__(self, "overrides")
        self._allow_system_aiml = False
        self._allow_learn_aiml = False
        self._allow_learnf_aiml = False

    @property
    def allow_system_aiml(self):
        return self._allow_system_aiml

    @property
    def allow_learn_aiml(self):
        return self._allow_learn_aiml

    @property
    def allow_learnf_aiml(self):
        return self._allow_learnf_aiml

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        overrides = configuration_file.get_section(self._section_name, configuration)
        if overrides is not None:
            self._allow_system_aiml = configuration_file.get_bool_option(overrides, "allow_system_aiml",
                                                                         missing_value=False, subs=subs)
            self._allow_learn_aiml = configuration_file.get_bool_option(overrides, "allow_learn_aiml",
                                                                        missing_value=False, subs=subs)
            self._allow_learnf_aiml = configuration_file.get_bool_option(overrides, "allow_learnf_aiml",
                                                                         missing_value=False, subs=subs)
        else:
            YLogger.warning(self, "'overrides' section missing from brain config, using to defaults")

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['allow_system_aiml'] = False
            data['allow_learn_aiml'] = False
            data['allow_learnf_aiml'] = False

        else:
            data['allow_system_aiml'] = self._allow_system_aiml
            data['allow_learn_aiml'] = self._allow_learn_aiml
            data['allow_learnf_aiml'] = self._allow_learnf_aiml



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class Timezones:

    RAW_DATA = [
        ["ACDT", "Australian Central Daylight Savings Time", "UTC+10:30"],
        ["ACST", "Australian Central Standard Time", "UTC+09:30"],
        ["ACT", "Acre Time", "UTC-05"],
        ["ACT", "ASEAN Common Time [unofficial]", "UTC+06:30 – UTC+09"],
        ["ACWST", "Australian Central Western Standard Time [unofficial]", "UTC+08:45"],
        ["ADT", "Atlantic Daylight Time", "UTC-03"],
        ["AEDT", "Australian Eastern Daylight Savings Time", "UTC+11"],
        ["AEST", "Australian Eastern Standard Time", "UTC+10"],
        ["AFT", "Afghanistan Time", "UTC+04:30"],
        ["AKDT", "Alaska Daylight Time", "UTC-08"],
        ["AKST", "Alaska Standard Time", "UTC-09"],
        ["ALMT", "Alma-Ata Time[1]", "UTC+06"],
        ["AMST", "Amazon Summer Time [Brazil][2]", "UTC-03"],
        ["AMT", "Amazon Time [Brazil][3]", "UTC-04"],
        ["AMT", "Armenia Time", "UTC+04"],
        ["ANAT", "Anadyr Time[4]", "UTC+12"],
        ["AQTT", "Aqtobe Time[5]", "UTC+05"],
        ["ART", "Argentina Time", "UTC-03"],
        ["AST", "Arabia Standard Time", "UTC+03"],
        ["AST", "Atlantic Standard Time", "UTC-04"],
        ["AWST", "Australian Western Standard Time", "UTC+08"],
        ["AZOST", "Azores Summer Time", "UTC±00"],
        ["AZOT", "Azores Standard Time", "UTC-01"],
        ["AZT", "Azerbaijan Time", "UTC+04"],
        ["BDT", "Brunei Time", "UTC+08"],
        ["BIOT", "British Indian Ocean Time", "UTC+06"],
        ["BIT", "Baker Island Time", "UTC-12"],
        ["BOT", "Bolivia Time", "UTC-04"],
        ["BRST", "Brasília Summer Time", "UTC-02"],
        ["BRT", "Brasília Time", "UTC-03"],
        ["BST", "Bangladesh Standard Time", "UTC+06"],
        ["BST", "Bougainville Standard Time[6]", "UTC+11"],
        ["BST", "British Summer Time [British Standard Time from Feb 1968 to Oct 1971]", "UTC+01"],
        ["BTT", "Bhutan Time", "UTC+06"],
        ["CAT", "Central Africa Time", "UTC+02"],
        ["CCT", "Cocos Islands Time", "UTC+06:30"],
        ["CDT", "Central Daylight Time [North America]", "UTC-05"],
        ["CDT", "Cuba Daylight Time[7]", "UTC-04"],
        ["CEST", "Central European Summer Time [Cf. HAEC]", "UTC+02"],
        ["CET", "Central European Time", "UTC+01"],
        ["CHADT", "Chatham Daylight Time", "UTC+13:45"],
        ["CHAST", "Chatham Standard Time", "UTC+12:45"],
        ["CHOT", "Choibalsan Standard Time", "UTC+08"],
        ["CHOST", "Choibalsan Summer Time", "UTC+09"],
        ["CHST", "Chamorro Standard Time", "UTC+10"],
        ["CHUT", "Chuuk Time", "UTC+10"],
        ["CIST", "Clipperton Island Standard Time", "UTC-08"],
        ["CIT", "Central Indonesia Time", "UTC+08"],
        ["CKT", "Cook Island Time", "UTC-10"],
        ["CLST", "Chile Summer Time", "UTC-03"],
        ["CLT", "Chile Standard Time", "UTC-04"],
        ["COST", "Colombia Summer Time", "UTC-04"],
        ["COT", "Colombia Time", "UTC-05"],
        ["CST", "Central Standard Time [North America]", "UTC-06"],
        ["CST", "China Standard Time", "UTC+08"],
        ["CST", "Cuba Standard Time", "UTC-05"],
        ["CT", "China Time", "UTC+08"],
        ["CVT", "Cape Verde Time", "UTC-01"],
        ["CWST", "Central Western Standard Time [Australia] unofficial", "UTC+08:45"],
        ["CXT", "Christmas Island Time", "UTC+07"],
        ["DAVT", "Davis Time", "UTC+07"],
        ["DDUT", "Dumont d'Urville Time", "UTC+10"],
        ["DFT", "AIX-specific equivalent of Central European Time[NB 1]", "UTC+01"],
        ["EASST", "Easter Island Summer Time", "UTC-05"],
        ["EAST", "Easter Island Standard Time", "UTC-06"],
        ["EAT", "East Africa Time", "UTC+03"],
        ["ECT", "Eastern Caribbean Time [does not recognise DST]", "UTC-04"],
        ["ECT", "Ecuador Time", "UTC-05"],
        ["EDT", "Eastern Daylight Time [North America]", "UTC-04"],
        ["EEST", "Eastern European Summer Time", "UTC+03"],
        ["EET", "Eastern European Time", "UTC+02"],
        ["EGST", "Eastern Greenland Summer Time", "UTC±00"],
        ["EGT", "Eastern Greenland Time", "UTC-01"],
        ["EIT", "Eastern Indonesian Time", "UTC+09"],
        ["EST", "Eastern Standard Time [North America]", "UTC-05"],
        ["FET", "Further-eastern European Time", "UTC+03"],
        ["FJT", "Fiji Time", "UTC+12"],
        ["FKST", "Falkland Islands Summer Time", "UTC-03"],
        ["FKT", "Falkland Islands Time", "UTC-04"],
        ["FNT", "Fernando de Noronha Time", "UTC-02"],
        ["GALT", "Galápagos Time", "UTC-06"],
        ["GAMT", "Gambier Islands Time", "UTC-09"],
        ["GET", "Georgia Standard Time", "UTC+04"],
        ["GFT", "French Guiana Time", "UTC-03"],
        ["GILT", "Gilbert Island Time", "UTC+12"],
        ["GIT", "Gambier Island Time", "UTC-09"],
        ["GMT", "Greenwich Mean Time", "UTC±00"],
        ["GST", "South Georgia and the South Sandwich Islands Time", "UTC-02"],
        ["GST", "Gulf Standard Time", "UTC+04"],
        ["GYT", "Guyana Time", "UTC-04"],
        ["HDT", "Hawaii–Aleutian Daylight Time", "UTC-09"],
        ["HAEC", "Heure Avancée d'Europe Centrale French-language name for CEST", "UTC+02"],
        ["HST", "Hawaii–Aleutian Standard Time", "UTC-10"],
        ["HKT", "Hong Kong Time", "UTC+08"],
        ["HMT", "Heard and McDonald Islands Time", "UTC+05"],
        ["HOVST", "Hovd Summer Time [not used from 2017-present]", "UTC+08"],
        ["HOVT", "Hovd Time", "UTC+07"],
        ["ICT", "Indochina Time", "UTC+07"],
        ["IDLW", "International Day Line West time zone", "UTC-12"],
        ["IDT", "Israel Daylight Time", "UTC+03"],
        ["IOT", "Indian Ocean Time", "UTC+03"],
        ["IRDT", "Iran Daylight Time", "UTC+04:30"],
        ["IRKT", "Irkutsk Time", "UTC+08"],
        ["IRST", "Iran Standard Time", "UTC+03:30"],
        ["IST", "Indian Standard Time", "UTC+05:30"],
        ["IST", "Irish Standard Time[8]", "UTC+01"],
        ["IST", "Israel Standard Time", "UTC+02"],
        ["JST", "Japan Standard Time", "UTC+09"],
        ["KALT", "Kaliningrad Time", "UTC+02"],
        ["KGT", "Kyrgyzstan Time", "UTC+06"],
        ["KOST", "Kosrae Time", "UTC+11"],
        ["KRAT", "Krasnoyarsk Time", "UTC+07"],
        ["KST", "Korea Standard Time", "UTC+09"],
        ["LHST", "Lord Howe Standard Time", "UTC+10:30"],
        ["LHST", "Lord Howe Summer Time", "UTC+11"],
        ["LINT", "Line Islands Time", "UTC+14"],
        ["MAGT", "Magadan Time", "UTC+12"],
        ["MART", "Marquesas Islands Time", "UTC-09:30"],
        ["MAWT", "Mawson Station Time", "UTC+05"],
        ["MDT", "Mountain Daylight Time [North America]", "UTC-06"],
        ["MET", "Middle European Time Same zone as CET", "UTC+01"],
        ["MEST", "Middle European Summer Time Same zone as CEST", "UTC+02"],
        ["MHT", "Marshall Islands Time", "UTC+12"],
        ["MIST", "Macquarie Island Station Time", "UTC+11"],
        ["MIT", "Marquesas Islands Time", "UTC-09:30"],
        ["MMT", "Myanmar Standard Time", "UTC+06:30"],
        ["MSK", "Moscow Time", "UTC+03"],
        ["MST", "Malaysia Standard Time", "UTC+08"],
        ["MST", "Mountain Standard Time [North America]", "UTC-07"],
        ["MUT", "Mauritius Time", "UTC+04"],
        ["MVT", "Maldives Time", "UTC+05"],
        ["MYT", "Malaysia Time", "UTC+08"],
        ["NCT", "New Caledonia Time", "UTC+11"],
        ["NDT", "Newfoundland Daylight Time", "UTC-02:30"],
        ["NFT", "Norfolk Island Time", "UTC+11"],
        ["NOVT", "Novosibirsk Time [9]", "UTC+07"],
        ["NPT", "Nepal Time", "UTC+05:45"],
        ["NST", "Newfoundland Standard Time", "UTC-03:30"],
        ["NT", "Newfoundland Time", "UTC-03:30"],
        ["NUT", "Niue Time", "UTC-11"],
        ["NZDT", "New Zealand Daylight Time", "UTC+13"],
        ["NZST", "New Zealand Standard Time", "UTC+12"],
        ["OMST", "Omsk Time", "UTC+06"],
        ["ORAT", "Oral Time", "UTC+05"],
        ["PDT", "Pacific Daylight Time [North America]", "UTC-07"],
        ["PET", "Peru Time", "UTC-05"],
        ["PETT", "Kamchatka Time", "UTC+12"],
        ["PGT", "Papua New Guinea Time", "UTC+10"],
        ["PHOT", "Phoenix Island Time", "UTC+13"],
        ["PHT", "Philippine Time", "UTC+08"],
        ["PKT", "Pakistan Standard Time", "UTC+05"],
        ["PMDT", "Saint Pierre and Miquelon Daylight Time", "UTC-02"],
        ["PMST", "Saint Pierre and Miquelon Standard Time", "UTC-03"],
        ["PONT", "Pohnpei Standard Time", "UTC+11"],
        ["PST", "Pacific Standard Time [North America]", "UTC-08"],
        ["PST", "Philippine Standard Time", "UTC+08"],
        ["PYST", "Paraguay Summer Time[10]", "UTC-03"],
        ["PYT", "Paraguay Time[11]", "UTC-04"],
        ["RET", "Réunion Time", "UTC+04"],
        ["ROTT", "Rothera Research Station Time", "UTC-03"],
        ["SAKT", "Sakhalin Island Time", "UTC+11"],
        ["SAMT", "Samara Time", "UTC+04"],
        ["SAST", "South African Standard Time", "UTC+02"],
        ["SBT", "Solomon Islands Time", "UTC+11"],
        ["SCT", "Seychelles Time", "UTC+04"],
        ["SDT", "Samoa Daylight Time", "UTC-10"],
        ["SGT", "Singapore Time", "UTC+08"],
        ["SLST", "Sri Lanka Standard Time", "UTC+05:30"],
        ["SRET", "Srednekolymsk Time", "UTC+11"],
        ["SRT", "Suriname Time", "UTC-03"],
        ["SST", "Samoa Standard Time", "UTC-11"],
        ["SST", "Singapore Standard Time", "UTC+08"],
        ["SYOT", "Showa Station Time", "UTC+03"],
        ["TAHT", "Tahiti Time", "UTC-10"],
        ["THA", "Thailand Standard Time", "UTC+07"],
        ["TFT", "French Southern and Antarctic Time[12]", "UTC+05"],
        ["TJT", "Tajikistan Time", "UTC+05"],
        ["TKT", "Tokelau Time", "UTC+13"],
        ["TLT", "Timor Leste Time", "UTC+09"],
        ["TMT", "Turkmenistan Time", "UTC+05"],
        ["TRT", "Turkey Time", "UTC+03"],
        ["TOT", "Tonga Time", "UTC+13"],
        ["TVT", "Tuvalu Time", "UTC+12"],
        ["ULAST", "Ulaanbaatar Summer Time", "UTC+09"],
        ["ULAT", "Ulaanbaatar Standard Time", "UTC+08"],
        ["UTC", "Coordinated Universal Time", "UTC±00"],
        ["UYST", "Uruguay Summer Time", "UTC-02"],
        ["UYT", "Uruguay Standard Time", "UTC-03"],
        ["UZT", "Uzbekistan Time", "UTC+05"],
        ["VET", "Venezuelan Standard Time", "UTC-04"],
        ["VLAT", "Vladivostok Time", "UTC+10"],
        ["VOLT", "Volgograd Time", "UTC+04"],
        ["VOST", "Vostok Station Time", "UTC+06"],
        ["VUT", "Vanuatu Time", "UTC+11"],
        ["WAKT", "Wake Island Time", "UTC+12"],
        ["WAST", "West Africa Summer Time", "UTC+02"],
        ["WAT", "West Africa Time", "UTC+01"],
        ["WEST", "Western European Summer Time", "UTC+01"],
        ["WET", "Western European Time", "UTC±00"],
        ["WIT", "Western Indonesian Time", "UTC+07"],
        ["WST", "Western Standard Time", "UTC+08"],
        ["YAKT", "Yakutsk Time", "UTC+09"],
        ["YEKT", "Yekaterinburg Time", "UTC+05"]
    ]

    CODES = {timezone[0]: timezone for timezone in RAW_DATA}

    TIMEZONES = {timezone[1]: timezone for timezone in RAW_DATA}



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
from datetime import datetime
try:
    import wikipedia

except ModuleNotFoundError as error:
    print("First use 'pip install wikipedia' before using this service")

from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.library.base import PythonAPIService


class WikipediaSearchQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return WikipediaSearchQuery(service)

    def parse_matched(self, matched):
        if len(matched) == 1:
            self._title = matched[0]

        else:
            raise ValueError("title missing")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._title = None

    def execute(self):
        return self._service.search(self._title)

    def aiml_response(self, response):
        result = "SEARCH <ul>" + "".join("<li>"+x+"</li>" for x in response['response']['payload']['search']) + "</ul>"
        YLogger.debug(self, result)
        return result


class WikipediaSummaryQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return WikipediaSummaryQuery(service)

    def parse_matched(self, matched):
        if len(matched) == 1:
            self._query = matched[0]

        else:
            raise ValueError("title missing")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._query = None

    def execute(self):
        return self._service.summary(self._query)

    def aiml_response(self, response):
        result = "SUMMARY {0}".format(response['response']['payload']['summary'])
        YLogger.debug(self, result)
        return result


class WikipediaService(PythonAPIService):

    PATTERNS = [
        [r"SEARCH\s(\w+)", WikipediaSearchQuery],
        [r"SUMMARY\s(.+)", WikipediaSummaryQuery]
    ]

    def __init__(self, configuration):
        PythonAPIService.__init__(self, configuration)

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "wikipedia.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "wikipedia.conf"

    def _response_to_json(self, api, response):
        return response

    # Provide a summary of a single article
    def summary(self, title, sentences=0, chars=0, auto_suggest=True, redirect=True):
        started = datetime.now()
        speed = None
        try:
            data = wikipedia.summary(title, sentences, chars, auto_suggest, redirect)
            speed = started - datetime.now()

            if data is not None:
                result = {"summary": data}
                return self._create_success_payload("summary", started, speed, result)

            return self._create_failure_payload("summary", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("summary", started, speed, error)

    # Provide a list of articles matching the query
    # Use summary to return the neccassary action
    def search(self, query, results=10, suggestion=False):
        started = datetime.now()
        speed = None
        try:
            data = wikipedia.search(query, results, suggestion)
            speed = started - datetime.now()

            if data is not None:
                result = {"search": data}
                return self._create_success_payload("search", started, speed, result)

            return self._create_failure_payload("search", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("search", started, speed, error)

    def patterns(self) -> list:
        return WikipediaService.PATTERNS



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class CountryCodes:

    RAW_DATA = [
        ["Afghanistan", "AF", "AFG", "4"],
        ["Albania", "AL", "ALB", "8"],
        ["Algeria", "DZ", "DZA", "12"],
        ["American Samoa", "AS", "ASM", "16"],
        ["Andorra", "AD", "AND", "20"],
        ["Angola", "AO", "AGO", "24"],
        ["Anguilla", "AI", "AIA", "660"],
        ["Antarctica", "AQ", "ATA", "10"],
        ["Antigua and Barbuda", "AG", "ATG", "28"],
        ["Argentina", "AR", "ARG", "32"],
        ["Armenia", "AM", "ARM", "51"],
        ["Aruba", "AW", "ABW", "533"],
        ["Australia", "AU", "AUS", "36"],
        ["Austria", "AT", "AUT", "40"],
        ["Azerbaijan", "AZ", "AZE", "31"],
        ["Bahamas", "BS", "BHS", "44"],
        ["Bahrain", "BH", "BHR", "48"],
        ["Bangladesh", "BD", "BGD", "50"],
        ["Barbados", "BB", "BRB", "52"],
        ["Belarus", "BY", "BLR", "112"],
        ["Belgium", "BE", "BEL", "56"],
        ["Belize", "BZ", "BLZ", "84"],
        ["Benin", "BJ", "BEN", "204"],
        ["Bermuda", "BM", "BMU", "60"],
        ["Bhutan", "BT", "BTN", "64"],
        ["Bolivia ", "BO", "BOL", "68"],
        ["Bonaire", "BQ", "BES", "535"],
        ["Sint Eustatius", "BQ", "BES", "535"],
        ["Saba", "BQ", "BES", "535"],
        ["Bosnia and Herzegovina", "BA", "BIH", "70"],
        ["Botswana", "BW", "BWA", "72"],
        ["Bouvet Island", "BV", "BVT", "74"],
        ["Brazil", "BR", "BRA", "76"],
        ["British Indian Ocean Territory", "IO", "IOT", "86"],
        ["Brunei Darussalam", "BN", "BRN", "96"],
        ["Bulgaria", "BG", "BGR", "100"],
        ["Burkina Faso", "BF", "BFA", "854"],
        ["Burundi", "BI", "BDI", "108"],
        ["Cabo Verde", "CV", "CPV", "132"],
        ["Cambodia", "KH", "KHM", "116"],
        ["Cameroon", "CM", "CMR", "120"],
        ["Canada", "CA", "CAN", "124"],
        ["Cayman Islands", "KY", "CYM", "136"],
        ["Central African Republic", "CF", "CAF", "140"],
        ["Chad", "TD", "TCD", "148"],
        ["Chile", "CL", "CHL", "152"],
        ["China", "CN", "CHN", "156"],
        ["Christmas Island", "CX", "CXR", "162"],
        ["Cocos [Keeling] Islands", "CC", "CCK", "166"],
        ["Colombia", "CO", "COL", "170"],
        ["Comoros", "KM", "COM", "174"],
        ["Congo", "CD", "COD", "180"],
        ["Congo", "CG", "COG", "178"],
        ["Cook Islands", "CK", "COK", "184"],
        ["Costa Rica", "CR", "CRI", "188"],
        ["Croatia", "HR", "HRV", "191"],
        ["Cuba", "CU", "CUB", "192"],
        ["Cura�ao", "CW", "CUW", "531"],
        ["Cyprus", "CY", "CYP", "196"],
        ["Czechia", "CZ", "CZE", "203"],
        ["C�te d'Ivoire", "CI", "CIV", "384"],
        ["Denmark", "DK", "DNK", "208"],
        ["Djibouti", "DJ", "DJI", "262"],
        ["Dominica", "DM", "DMA", "212"],
        ["Dominican Republic", "DO", "DOM", "214"],
        ["Ecuador", "EC", "ECU", "218"],
        ["Egypt", "EG", "EGY", "818"],
        ["El Salvador", "SV", "SLV", "222"],
        ["Equatorial Guinea", "GQ", "GNQ", "226"],
        ["Eritrea", "ER", "ERI", "232"],
        ["Estonia", "EE", "EST", "233"],
        ["Eswatini", "SZ", "SWZ", "748"],
        ["Ethiopia", "ET", "ETH", "231"],
        ["Falkland Islands", "FK", "FLK", "238"],
        ["Faroe Islands", "FO", "FRO", "234"],
        ["Fiji", "FJ", "FJI", "242"],
        ["Finland", "FI", "FIN", "246"],
        ["France", "FR", "FRA", "250"],
        ["French Guiana", "GF", "GUF", "254"],
        ["French Polynesia", "PF", "PYF", "258"],
        ["French Southern Territories", "TF", "ATF", "260"],
        ["Gabon", "GA", "GAB", "266"],
        ["Gambia", "GM", "GMB", "270"],
        ["Georgia", "GE", "GEO", "268"],
        ["Germany", "DE", "DEU", "276"],
        ["Ghana", "GH", "GHA", "288"],
        ["Gibraltar", "GI", "GIB", "292"],
        ["Greece", "GR", "GRC", "300"],
        ["Greenland", "GL", "GRL", "304"],
        ["Grenada", "GD", "GRD", "308"],
        ["Guadeloupe", "GP", "GLP", "312"],
        ["Guam", "GU", "GUM", "316"],
        ["Guatemala", "GT", "GTM", "320"],
        ["Guernsey", "GG", "GGY", "831"],
        ["Guinea", "GN", "GIN", "324"],
        ["Guinea-Bissau", "GW", "GNB", "624"],
        ["Guyana", "GY", "GUY", "328"],
        ["Haiti", "HT", "HTI", "332"],
        ["Heard Island and McDonald Islands", "HM", "HMD", "334"],
        ["Holy See", "VA", "VAT", "336"],
        ["Honduras", "HN", "HND", "340"],
        ["Hong Kong", "HK", "HKG", "344"],
        ["Hungary", "HU", "HUN", "348"],
        ["Iceland", "IS", "ISL", "352"],
        ["India", "IN", "IND", "356"],
        ["Indonesia", "ID", "IDN", "360"],
        ["Iran", "IR", "IRN", "364"],
        ["Iraq", "IQ", "IRQ", "368"],
        ["Ireland", "IE", "IRL", "372"],
        ["Isle of Man", "IM", "IMN", "833"],
        ["Israel", "IL", "ISR", "376"],
        ["Italy", "IT", "ITA", "380"],
        ["Jamaica", "JM", "JAM", "388"],
        ["Japan", "JP", "JPN", "392"],
        ["Jersey", "JE", "JEY", "832"],
        ["Jordan", "JO", "JOR", "400"],
        ["Kazakhstan", "KZ", "KAZ", "398"],
        ["Kenya", "KE", "KEN", "404"],
        ["Kiribati", "KI", "KIR", "296"],
        ["Noth Korea", "KP", "PRK", "408"],
        ["South Korea", "KR", "KOR", "410"],
        ["Kuwait", "KW", "KWT", "414"],
        ["Kyrgyzstan", "KG", "KGZ", "417"],
        ["Lao People's Democratic Republic", "LA", "LAO", "418"],
        ["Latvia", "LV", "LVA", "428"],
        ["Lebanon", "LB", "LBN", "422"],
        ["Lesotho", "LS", "LSO", "426"],
        ["Liberia", "LR", "LBR", "430"],
        ["Libya", "LY", "LBY", "434"],
        ["Liechtenstein", "LI", "LIE", "438"],
        ["Lithuania", "LT", "LTU", "440"],
        ["Luxembourg", "LU", "LUX", "442"],
        ["Macao", "MO", "MAC", "446"],
        ["Madagascar", "MG", "MDG", "450"],
        ["Malawi", "MW", "MWI", "454"],
        ["Malaysia", "MY", "MYS", "458"],
        ["Maldives", "MV", "MDV", "462"],
        ["Mali", "ML", "MLI", "466"],
        ["Malta", "MT", "MLT", "470"],
        ["Marshall Islands", "MH", "MHL", "584"],
        ["Martinique", "MQ", "MTQ", "474"],
        ["Mauritania", "MR", "MRT", "478"],
        ["Mauritius", "MU", "MUS", "480"],
        ["Mayotte", "YT", "MYT", "175"],
        ["Mexico", "MX", "MEX", "484"],
        ["Micronesia", "FM", "FSM", "583"],
        ["Moldova", "MD", "MDA", "498"],
        ["Monaco", "MC", "MCO", "492"],
        ["Mongolia", "MN", "MNG", "496"],
        ["Montenegro", "ME", "MNE", "499"],
        ["Montserrat", "MS", "MSR", "500"],
        ["Morocco", "MA", "MAR", "504"],
        ["Mozambique", "MZ", "MOZ", "508"],
        ["Myanmar", "MM", "MMR", "104"],
        ["Namibia", "NA", "NAM", "516"],
        ["Nauru", "NR", "NRU", "520"],
        ["Nepal", "NP", "NPL", "524"],
        ["Netherlands", "NL", "NLD", "528"],
        ["New Caledonia", "NC", "NCL", "540"],
        ["New Zealand", "NZ", "NZL", "554"],
        ["Nicaragua", "NI", "NIC", "558"],
        ["Niger", "NE", "NER", "562"],
        ["Nigeria", "NG", "NGA", "566"],
        ["Niue", "NU", "NIU", "570"],
        ["Norfolk Island", "NF", "NFK", "574"],
        ["Northern Mariana Islands", "MP", "MNP", "580"],
        ["Norway", "NO", "NOR", "578"],
        ["Oman", "OM", "OMN", "512"],
        ["Pakistan", "PK", "PAK", "586"],
        ["Palau", "PW", "PLW", "585"],
        ["Palestine", "PS", "PSE", "275"],
        ["Panama", "PA", "PAN", "591"],
        ["Papua New Guinea", "PG", "PNG", "598"],
        ["Paraguay", "PY", "PRY", "600"],
        ["Peru", "PE", "PER", "604"],
        ["Philippines", "PH", "PHL", "608"],
        ["Pitcairn", "PN", "PCN", "612"],
        ["Poland", "PL", "POL", "616"],
        ["Portugal", "PT", "PRT", "620"],
        ["Puerto Rico", "PR", "PRI", "630"],
        ["Qatar", "QA", "QAT", "634"],
        ["Republic of North Macedonia", "MK", "MKD", "807"],
        ["Romania", "RO", "ROU", "642"],
        ["Russian Federation", "RU", "RUS", "643"],
        ["Rwanda", "RW", "RWA", "646"],
        ["R�union", "RE", "REU", "638"],
        ["Saint Barth�lemy", "BL", "BLM", "652"],
        ["Saint Helena", "SH", "SHN", "654"],
        ["Ascension", "SH", "SHN", "654"],
        ["Tristan da Cunha", "SH", "SHN", "654"],
        ["Saint Kitts and Nevis", "KN", "KNA", "659"],
        ["Saint Lucia", "LC", "LCA", "662"],
        ["Saint Martin", "MF", "MAF", "663"],
        ["Saint Pierre and Miquelon", "PM", "SPM", "666"],
        ["Saint Vincent and the Grenadines", "VC", "VCT", "670"],
        ["Samoa", "WS", "WSM", "882"],
        ["San Marino", "SM", "SMR", "674"],
        ["Sao Tome and Principe", "ST", "STP", "678"],
        ["Saudi Arabia", "SA", "SAU", "682"],
        ["Senegal", "SN", "SEN", "686"],
        ["Serbia", "RS", "SRB", "688"],
        ["Seychelles", "SC", "SYC", "690"],
        ["Sierra Leone", "SL", "SLE", "694"],
        ["Singapore", "SG", "SGP", "702"],
        ["Dutch Sint Maarten", "SX", "SXM", "534"],
        ["Slovakia", "SK", "SVK", "703"],
        ["Slovenia", "SI", "SVN", "705"],
        ["Solomon Islands", "SB", "SLB", "90"],
        ["Somalia", "SO", "SOM", "706"],
        ["South Africa", "ZA", "ZAF", "710"],
        ["South Georgia and the South Sandwich Islands", "GS", "SGS", "239"],
        ["South Sudan", "SS", "SSD", "728"],
        ["Spain", "ES", "ESP", "724"],
        ["Sri Lanka", "LK", "LKA", "144"],
        ["Sudan", "SD", "SDN", "729"],
        ["Suriname", "SR", "SUR", "740"],
        ["Svalbard and Jan Mayen", "SJ", "SJM", "744"],
        ["Sweden", "SE", "SWE", "752"],
        ["Switzerland", "CH", "CHE", "756"],
        ["Syrian Arab Republic", "SY", "SYR", "760"],
        ["Taiwan", "TW", "TWN", "158"],
        ["Tajikistan", "TJ", "TJK", "762"],
        ["Tanzania", "TZ", "TZA", "834"],
        ["Thailand", "TH", "THA", "764"],
        ["Timor-Leste", "TL", "TLS", "626"],
        ["Togo", "TG", "TGO", "768"],
        ["Tokelau", "TK", "TKL", "772"],
        ["Tonga", "TO", "TON", "776"],
        ["Trinidad and Tobago", "TT", "TTO", "780"],
        ["Tunisia", "TN", "TUN", "788"],
        ["Turkey", "TR", "TUR", "792"],
        ["Turkmenistan", "TM", "TKM", "795"],
        ["Turks and Caicos Islands", "TC", "TCA", "796"],
        ["Tuvalu", "TV", "TUV", "798"],
        ["Uganda", "UG", "UGA", "800"],
        ["Ukraine", "UA", "UKR", "804"],
        ["United Arab Emirates", "AE", "ARE", "784"],
        ["United Kingdom", "GB", "GBR", "826"],
        ["Great Britain", "GB", "GBR", "826"],
        ["United Kingdom of Great Britain and Northern Ireland", "GB", "GBR", "826"],
        ["United States Minor Outlying Islands", "UM", "UMI", "581"],
        ["United States of America", "US", "USA", "840"],
        ["Uruguay", "UY", "URY", "858"],
        ["Uzbekistan", "UZ", "UZB", "860"],
        ["Vanuatu", "VU", "VUT", "548"],
        ["Venezuela", "VE", "VEN", "862"],
        ["Viet Nam", "VN", "VNM", "704"],
        ["British Virgin Islands", "VG", "VGB", "92"],
        ["US Virgin Islands", "VI", "VIR", "850"],
        ["Wallis and Futuna", "WF", "WLF", "876"],
        ["Western Sahara", "EH", "ESH", "732"],
        ["Yemen", "YE", "YEM", "887"],
        ["Zambia", "ZM", "ZMB", "894"],
        ["Zimbabwe", "ZW", "ZWE", "716"],
        ["�land Islands", "AX", "ALA", "248"],
    ]
    
    NAMES = {country[0]: country for country in RAW_DATA}

    TWO_DIGITS = {country[1]: country for country in RAW_DATA}

    THREE_DIGITS = {country[2]: country for country in RAW_DATA}

    NUMERIC = {country[3]: country for country in RAW_DATA}



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import urllib.parse
import os
from programy.services.base import ServiceQuery
from programy.services.rest.base import RESTService
from programy.services.rest.base import RESTServiceException
from programy.utils.logging.ylogger import YLogger


class DuckDuckInstantQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return DuckDuckInstantQuery(service)

    def parse_matched(self, matched):
        self._query = ServiceQuery._get_matched_var(matched, 0, "query")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._query = None

    def execute(self):
        return self._service.instant(self._query)

    def aiml_response(self, response):
        payload = response['response']['payload']
        result = payload['AbstractText']
        YLogger.debug(self, result)
        return result


class DuckDuckScrapeQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return DuckDuckInstantQuery(service)

    def parse_matched(self, matched):
        self._query = ServiceQuery._get_matched_var(matched, 0, "query")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._query = None

    def execute(self):
        return self._service.scrape(self._query)

    def aiml_response(self, response):
        payload = response['response']['payload']
        result = payload['AbstractText']
        YLogger.debug(self, result)
        return result


class DuckDuckGoServiceException(RESTServiceException):

    def __init__(self, msg):
        RESTServiceException.__init__(self, msg)


class DuckDuckGoService(RESTService):
    """
    http://www.duckduckgo.com
    """
    PATTERNS = [
        [r"INSTANT\s(.+)", DuckDuckInstantQuery],
        [r"SCRAPE\s(.+)", DuckDuckScrapeQuery]
    ]

    BASE_INSTANT_URL="http://api.duckduckgo.com"
    BASE_HTTPSCRAPE_URL="https://duckduckgo.com/html/q={0}"

    def __init__(self, configuration):
        RESTService.__init__(self, configuration)
        self._account_name = None
        self._country = None

    def patterns(self) -> list:
        return DuckDuckGoService.PATTERNS

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "duckduckgo.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "duckduckgo.conf"

    def _build_instant_url(self, term):
        url = DuckDuckGoService.BASE_INSTANT_URL
        url += "?q={0}".format(urllib.parse.quote(term))
        url += "&format=json"
        return url

    def instant(self, term):
        url = self._build_instant_url(term)
        response = self.query('instant', url)
        return response

    def _build_scrape_url(self, term):
        url = DuckDuckGoService.BASE_HTTPSCRAPE_URL.format(urllib.parse.quote(term))
        return url

    def scrape(self, term):
        url = self._build_instant_url(term)
        response = self.query('scrape', url)
        return response

    def _response_to_json(self, api, response):
        return response.json()
